#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "view.h"
#include "display.h"
#include "object.h"

Object *init_object(char *datafilename, double xs, double ys, double zs)
{
   FILE           *datafile;
   double          minx;
   double          maxx;
   double          miny;
   double          maxy;
   double          minz;
   double          maxz;
   int             i;
   int             Index;
   int            *int_Pointer;
   Object         *object;
   int             nvertex;
   vertex         *vertices;
   int             npoly;
   int           **poly_list;
   vertex       ***poly_vertices;
   vertex        **double_Pointer;

   if ((datafile = fopen(datafilename, "r")) == NULL)
   {
      fprintf(stderr, "Error opening %s\n", datafile);
      exit(1);
   }

   if ((object = (Object *) malloc(sizeof(Object))) == NULL)
   {
      perror("init_object()");
      exit(1);
   }

   minx = miny = minz = HUGE_VAL;
   maxx = maxy = maxz = -HUGE_VAL;

   fscanf(datafile, "%d", &nvertex);
   object->nvertex = nvertex;

   if ((vertices = (vertex *) malloc(sizeof(vertex) * nvertex)) == NULL)
   {
      perror("init_object()");
      exit(1);
   }
   object->vertices = vertices;

   for (i = 0; i < nvertex; i++)
   {
      /*
       * switch y's and z's.
       */
      fscanf(datafile, "%lf %lf %lf", &(vertices[i].Coordinate[X]),
             &(vertices[i].Coordinate[Y]),
             &(vertices[i].Coordinate[Z]));
      vertices[i].Coordinate[X] *= xs;
      vertices[i].Coordinate[Y] *= ys;
      vertices[i].Coordinate[Z] *= zs;

      minx = MIN(vertices[i].Coordinate[X], minx);
      miny = MIN(vertices[i].Coordinate[Y], miny);
      minz = MIN(vertices[i].Coordinate[Z], minz);

      maxx = MAX(vertices[i].Coordinate[X], maxx);
      maxy = MAX(vertices[i].Coordinate[Y], maxy);
      maxz = MAX(vertices[i].Coordinate[Z], maxz);
   }

   fscanf(datafile, "%d", &npoly);
   object->npoly = npoly;

   if ((poly_list = (int **) malloc(sizeof(int *) * npoly)) == NULL)
   {
      perror("init_object()");
      exit(1);
   }
   object->poly_list = poly_list;

   if ((poly_vertices = (vertex ***) malloc(sizeof(vertex **) * npoly))
       == NULL)
   {
      perror("init_object()");
      exit(1);
   }
   object->poly_vertices = poly_vertices;

   if ((double_Pointer = (vertex **) malloc(sizeof(vertex *) * 3 * npoly))
       == NULL)
   {
      perror("init_object()");
      exit(1);
   }

   if ((int_Pointer = (int *) malloc(sizeof(int) * 3 * npoly)) == NULL)
   {
      perror("init_object()");
      exit(1);
   }

   for (i = 0, Index = 0; i < npoly; i++, Index += 3)
   {
      poly_list[i] = &(int_Pointer[Index]);

      poly_vertices[i] = &(double_Pointer[Index]);

      fscanf(datafile, "%d %d %d", &(poly_list[i][0]), &(poly_list[i][1]),
             &(poly_list[i][2]));

      poly_vertices[i][0] = &(vertices[poly_list[i][0]]);
      poly_vertices[i][1] = &(vertices[poly_list[i][1]]);
      poly_vertices[i][2] = &(vertices[poly_list[i][2]]);
   }

   fclose(datafile);

   /*
    * printf( "done reading: # vertices= %d, # polys = %d \n",
    * nvertex,npoly);
    */
   printf("minimum: <%f %f %f>\n", minx, miny, minz);
   printf("maximum: <%f %f %f>\n", maxx, maxy, maxz);

   object ->Min_point.x = minx;
   object ->Min_point.y = miny;
   object ->Min_point.z = minz;

   object ->Max_point.x = maxx;
   object ->Max_point.y = maxy;
   object ->Max_point.z = maxz;

   return(object);
}

void transform_object(Transformation  *View, Object *object)
{
   register double *Matrix_row_0 = (View->T_matrix)[0];
   register double *Matrix_row_1 = (View->T_matrix)[1];
   register double *Matrix_row_2 = (View->T_matrix)[2];
   register double *Matrix_row_3 = (View->T_matrix)[3];

   register int     nvertex         = object->nvertex;
   register vertex *vertices        = object->vertices;
   register double *next_vertex;

   register int     i;

   double           x_hat;
   double           y_hat;
   double           z_hat;

   for (i = nvertex - 1; i >= 0; i--)
   {
      x_hat = vertices[i].Coordinate[X];
      y_hat = vertices[i].Coordinate[Y];
      z_hat = vertices[i].Coordinate[Z];

      next_vertex = vertices[i].Transformed_Coordinate;

      next_vertex[0] = Matrix_row_0[0] * x_hat + Matrix_row_0[1] * y_hat
                       + Matrix_row_0[2] * z_hat + Matrix_row_0[3];

      next_vertex[1] = Matrix_row_1[0] * x_hat + Matrix_row_1[1] * y_hat
                       + Matrix_row_1[2] * z_hat + Matrix_row_1[3];

      next_vertex[2] = Matrix_row_2[0] * x_hat + Matrix_row_2[1] * y_hat
                       + Matrix_row_2[2] * z_hat + Matrix_row_2[3];

      next_vertex[3] = Matrix_row_3[0] * x_hat + Matrix_row_3[1] * y_hat
                       + Matrix_row_3[2] * z_hat + Matrix_row_3[3];
   }
}

void            draw_object(Viewer *viewer, Transformation *View,
                            Object *object)
{
   register int       npoly         = object->npoly;
   register vertex ***poly_vertices = object->poly_vertices;
   register vertex  **next_poly;

   Display *theDisplay         = viewer->theDisplay;
   Window   theDA              = viewer->theDA;
   GC       theGC              = viewer->theGC;
   long     theBackgroundPixel = viewer->theBackgroundPixel;
   int      height             = viewer->height;
   int      width              = viewer->width;
   Pixmap   thePixmap          = viewer->thePixmap;

   register int    i;


   register vertex         *v0;
   register vertex         *v1;
   register vertex         *v2;

   XSetForeground(theDisplay, theGC, theBackgroundPixel);
   XFillRectangle(theDisplay, thePixmap, theGC, 0, 0, width, height);

   for (i = 0; i < npoly; i++)
   {
      next_poly = poly_vertices[i];

      v0 = next_poly[0];
      v1 = next_poly[1];
      v2 = next_poly[2];

      draw_homogenize_line(viewer, View, v0, v1);
      draw_homogenize_line(viewer, View, v1, v2);
      draw_homogenize_line(viewer, View, v2, v0);
   }
   XCopyArea(theDisplay, thePixmap, theDA, theGC, 0, 0, width, height, 0, 0);
}

void draw_homogenize_line(Viewer *viewer, Transformation *View,
                          vertex *v0, vertex *v1)
{
   Display *theDisplay         = viewer->theDisplay;
   GC       theGC              = viewer->theGC;
   long     theForegroundPixel = viewer->theForegroundPixel;
   Pixmap   thePixmap          = viewer->thePixmap;

   double   Zmin               = View->Zmin;

   vector line_vertex0;
   vector line_vertex1;
   int    copy_length = 4 * sizeof(double);

   memcpy(line_vertex0, v0->Transformed_Coordinate, copy_length);
   memcpy(line_vertex1, v1->Transformed_Coordinate, copy_length);

   if (clip_3d(line_vertex0, line_vertex1, Zmin))
   {
      homogenize_vector_transform(line_vertex0, View->vp_matrix);
      homogenize_vector_transform(line_vertex1, View->vp_matrix);

      XSetForeground(theDisplay, theGC, theForegroundPixel);
      XDrawLine(theDisplay, thePixmap, theGC,
                (int) line_vertex0[0], (int) line_vertex0[1],
                (int) line_vertex1[0], (int) line_vertex1[1]);
   }
}
